import ot
import jax
import numpy as np
import jax.numpy as jnp
from ott.solvers import linear
from ott.geometry import geometry

import jax
import jax.numpy as jnp
from jax.scipy.special import logsumexp

from uot.algorithms.utils import regularize_input

@jax.jit
def coupling_tensor(u: jnp.ndarray,
                    v: jnp.ndarray,
                    cost: jnp.ndarray,
                    epsilon: float) -> jnp.ndarray:
    return jnp.exp((u[:, None] + v[None, :] - cost) / epsilon)

@jax.jit
def tensor_marginals(matrix: jnp.ndarray):
    row_marginal = jnp.sum(matrix, axis=1)      # sum over columns
    column_marginal = jnp.sum(matrix, axis=0)   # sum over rows
    return (row_marginal, column_marginal)

@jax.jit
def compute_error(u: jnp.ndarray,
                  v: jnp.ndarray,
                  a: jnp.ndarray,
                  b: jnp.ndarray,
                  cost: jnp.ndarray,
                  epsilon: float) -> float:
    P = coupling_tensor(u, v, cost, epsilon)
    row_marginal, col_marginal = tensor_marginals(P)
    return jnp.max(
        jnp.array([
            jnp.linalg.norm(a - row_marginal),
            jnp.linalg.norm(b - col_marginal)
        ])
    ) 


@jax.jit
def sinkhorn(a: jnp.ndarray,
             b: jnp.ndarray,
             cost: jnp.ndarray,
             epsilon: float = 1e-3,
             precision: float = 1e-4,
             max_iters: int = 10_000):
    
    n = a.shape[0]
    m = b.shape[0]

    # Initialize dual variables
    u = jnp.zeros(n)
    v = jnp.zeros(m)

    # Define the loop condition
    def cond_fn(carry):
        u_, v_, i_, err_ = carry
        return jnp.logical_and(err_ > precision, i_ < max_iters)

    # Define one iteration body
    def body_fn(carry):
        u_, v_, i_, e = carry

        u_upd = u_ + epsilon * jnp.log(a) - epsilon * logsumexp((u_[:, None] + v_[None, :] - cost) / epsilon, axis=1)
        v_upd = v_ + epsilon * jnp.log(b) - epsilon * logsumexp((u_upd[:, None] + v_[None, :] - cost) / epsilon, axis=0)

        err_upd = jax.lax.cond(
            i_ % 10 == 0,
            lambda: compute_error(u_upd, v_upd, a, b, cost, epsilon),
            lambda: e
        )

        return (u_upd, v_upd, i_ + 1, err_upd)

    # Run the loop
    init_err = compute_error(u, v, a, b, cost, epsilon)
    u_final, v_final, i_final, final_err = jax.lax.while_loop(
        cond_fn,
        body_fn,
        (u, v, jnp.array(0), init_err)
    )

    return u_final, v_final


def jax_sinkhorn(a, b, C, epsilon=1e-3, tolerance = 1e-4):
    a, b, C = regularize_input(a, b, C)
    u, v = sinkhorn(a, b, C, epsilon)
    u.block_until_ready()
    v.block_until_ready()
    P = coupling_tensor(u, v, C, epsilon)

    error = compute_error(u, v, a, b, C, epsilon)
    converged = error < tolerance

    return P, jnp.sum(P * C), converged


def sink(a, b, cost, epsilon=1e-3):
    return linear.solve(
        geometry.Geometry(cost_matrix=cost, epsilon=epsilon),
        a=a,
        b=b,
        lse_mode=True,
        threshold=1e-4
    )

sink_2vmap = jax.jit(sink)

def ott_jax_sinkhorn(mu, nu, C, epsilon=0.001, threshold=1e-4):
    solution = sink_2vmap(mu, nu, C)
    solution.matrix.block_until_ready()
    return solution.matrix, jnp.sum(solution.matrix * C)



import jax
jax.config.update("jax_enable_x64", True)

import ot
import numpy as np
from tqdm import tqdm
from functools import partial

from sklearn.datasets import load_digits
from sklearn.preprocessing import normalize
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import make_pipeline
from sklearn.svm import SVC

from algorithms.sinkhorn import jax_sinkhorn
from algorithms.gradient_ascent import gradient_ascent
from algorithms.lbfgs import dual_lbfgs
from algorithms.lp import pot_lp


np.random.seed(32)

SAMPLE_SIZES = [100,
                200,
                300,
                500,
                1000]

digits = load_digits()
X, y = digits.data, digits.target 

X = normalize(X, axis=1)

solvers = {
    'sinkhorn': jax_sinkhorn,
    # 'grad-ascent': gradient_ascent,
    # 'dual-lbfs': dual_lbfgs,
    # 'lp': pot_lp
}

row, col = np.arange(8), np.arange(8)
row, col = np.meshgrid(row, col)
points = np.vstack([coordinate.ravel() for coordinate in [row, col]]).T
C = ot.dist(points, points)

pairwise_distances = {}

with tqdm(total=len(solvers) * len(X) ** 2, desc=f"Computing distances") as pbar:
    for solver_name, solver in solvers.items():
        distances_matrix = np.zeros((X.shape[0], X.shape[0]))
        pairwise_distances[solver_name] = distances_matrix

        for i in range(X.shape[0]):
            for j in range(X.shape[0]):
                if i > j:
                    continue
                distance = solver(X[i, :], X[j,:], C)[1]
                distances_matrix[i, j] = distance
                distances_matrix[j, i] = distance
                pbar.update(1)


def ot_kernel(X, Y, solver_name):
    kernel_matrix = np.zeros((X.shape[0], Y.shape[0]))
    distance_matrix = pairwise_distances[solver_name]
    for i, p in enumerate(X):
        for j, q in enumerate(Y):
            pbar.update(1)
            kernel_matrix[i, j] = np.exp(-distance_matrix[i, j])
    return kernel_matrix


for sample_size in SAMPLE_SIZES:
    indices = np.random.choice(X.shape[0], size=int(sample_size), replace=False)

    X_sub = X[indices] 
    y_sub = y[indices]

    kernel = lambda x, y: ot_kernel(x, y, solver_name="jax-sinkhorn")
    pipeline = make_pipeline(SVC(kernel=kernel, C=10, gamma=0.05))
    scores = cross_val_score(pipeline, X_sub, y_sub, cv=5, scoring='accuracy')


y = y.astype(int)



import jax
jax.config.update("jax_enable_x64", True)

import sys
import argparse
import numpy as np
import pandas as pd
from datetime import datetime
from uot.algorithms.sinkhorn import jax_sinkhorn
from uot.algorithms.gradient_ascent import gradient_ascent
from uot.algorithms.lbfgs import lbfgs_ot
from uot.algorithms.lp import pot_lp
from uot.core.experiment import run_experiment
from uot.core.analysis import get_agg_table
from uot.core.suites import time_precision_experiment

epsilon_kwargs = [
    # {'epsilon': 100},
    # {'epsilon': 10},
    {'epsilon': 1},
    # {'epsilon': 1e-1},
    # {'epsilon': 1e-3},
    # {'epsilon': 1e-6},
    # {'epsilon': 1e-9},
]

solvers = {
    'pot-lp': (pot_lp, []),
    'lbfgs': (lbfgs_ot, epsilon_kwargs),
    'jax-sinkhorn': (jax_sinkhorn, epsilon_kwargs),
    'grad-ascent': (gradient_ascent, epsilon_kwargs),
}

# algorithms that use jax jit 
jit_algorithms = [
    'jax-sinkhorn', 'optax-grad-ascent', 'lbfgs'
]

problemset_names = [
    (1, "gamma", 32),
    # (1, "gamma", 64),
    # (1, "gamma", 256),
    # (1, "gamma", 512),
    # (1, "gamma", 1024),
    # (1, "gamma", 2048),

    # (1, "gaussian", 32),
    # (1, "gaussian", 64),
    # (1, "gaussian", 256),
    # (1, "gaussian", 512),
    # (1, "gaussian", 1024),
    # (1, "gaussian", 2048),

    # (1, "beta", 32),
    # (1, "beta", 64),
    # (1, "beta", 256),
    # (1, "beta", 512),
    # (1, "beta", 1024),
    # (1, "beta", 2048),

    # (1, "gaussian|gamma|beta|cauchy", 32),
    # (1, "gaussian|gamma|beta|cauchy", 64),
    # (1, "gaussian|gamma|beta|cauchy", 128),
    # (1, "gaussian|gamma|beta|cauchy", 256),
    # (1, "gaussian|gamma|beta|cauchy", 512),
    # (1, "gaussian|gamma|beta|cauchy", 1024),
    # (1, "gaussian|gamma|beta|cauchy", 2048),

    # (2, "WhiteNoise", 32),
    # (2, "CauchyDensity", 32),
    # (2, "GRFmoderate", 32),
    # (2, "GRFrough", 32),
    # (2, "GRFsmooth", 32),
    # (2, "LogGRF", 32),
    # (2, "LogitGRF", 32),
    # (2, "MicroscopyImages", 32),
    # (2, "Shapes", 32),
    # (2, "ClassicImages", 64),

    # (3, "3dmesh", 1024),
    # (3, "3dmesh", 2048),
]

parser = argparse.ArgumentParser(description="Run time precision experiments.")
parser.add_argument(
    "--algorithms",
    nargs="+",
    default=list(solvers.keys()),
    help="List of algorithms to run experiments on. Use 'all' to run on all algorithms."
)
parser.add_argument(
    "--save",
    action="store_true",
    help="Flag to save the results to a CSV file."
)
parser.add_argument(
    "--show_agg",
    action="store_true",
    help="Flag to display the aggregated results table."
)
parser.add_argument(
    "--show",
    action="store_true",
    help="Flag to display the full results dataframe."
)
parser.add_argument(
    "--folds",
    type=int,
    default=5,
    help="Number of folds for cross-validation or repeated experiments."
)

args = parser.parse_args()

if "all" in args.algorithms:
    selected_solvers = solvers
else:
    selected_solvers = {key: solvers[key] for key in args.algorithms if key in solvers}


df = run_experiment(experiment=time_precision_experiment, 
                    problemsets_names=problemset_names,
                    solvers=selected_solvers,
                    jit_algorithms=jit_algorithms,
                    folds=args.folds)


if args.save:
    df.to_csv(f"results/result_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.csv")

if args.show:
    print(df)

if args.show_agg:
    dfs = [df[df.name == name] for name in df.name.unique()]

    for df in dfs:
        print("Solver", df.name.iloc[0])
        print(get_agg_table(df, ['time']))


import jax
import jax.numpy as jnp
import optax

def gradient_vanilla_logsumexp(
    a, b, C, eps=0.01, alpha=0.1, tolerance=1e-4, max_iterations=1_000
):
    """
    Vanilla gradient ascent for entropy-regularized optimal transport (Kantorovich problem).

    Arguments:
    C -- cost matrix (n x m)
    a -- source distribution (n,)
    b -- target distribution (m,)
    eps -- entropy regularization coefficient
    alpha -- initial learning rate
    tolerance -- stopping criterion for convergence
    max_iterations -- maximum number of iterations

    Returns:
    phi, psi -- dual potentials
    plan -- optimal transport plan
    """

    n, m = C.shape
    phi, psi = jnp.zeros(n), jnp.zeros(m)  # Dual variables initialization

    def compute_gradients(phi, psi):
        """Compute gradients using stabilized softmax approach."""
        log_K = (phi[:, None] + psi[None, :] - C) / eps  # Log of kernel
        log_K_max = jnp.max(log_K, axis=1, keepdims=True)  # Stabilization factor
        K_stable = jnp.exp(log_K - log_K_max)  # Stable exponentiation
        marginals = jnp.exp(log_K_max).flatten() * jnp.sum(
            K_stable, axis=1
        )  # Sum over columns
        grad_phi = a - marginals  # Compute gradient for phi

        log_K_max = jnp.max(log_K, axis=0, keepdims=True)
        K_stable = jnp.exp(log_K - log_K_max)
        marginals = jnp.exp(log_K_max).flatten() * jnp.sum(
            K_stable, axis=0
        )  # Sum over rows
        grad_psi = b - marginals  # Compute gradient for psi

        return grad_phi, grad_psi

    for iteration in range(max_iterations):
        grad_phi, grad_psi = compute_gradients(phi, psi)

        # Adaptive step size decay
        # step_size = alpha / (1 + 0.1 * iteration)
        step_size = alpha

        # Update dual variables
        phi_new = phi + step_size * grad_phi
        psi_new = psi + step_size * grad_psi

        # Convergence check
        # if (
        #     jnp.linalg.norm(grad_phi) < tolerance
        #     and jnp.linalg.norm(grad_psi) < tolerance
        # ):
        #     break
        if (
            jnp.linalg.norm(phi_new - phi) < tolerance
            and jnp.linalg.norm(psi_new - psi) < tolerance
        ):
            break
        phi = phi_new
        psi = psi_new

    phi = phi_new
    psi = psi_new

    # Compute final transport plan
    log_K = (phi[:, None] + psi[None, :] - C) / eps
    plan = jnp.exp(log_K) * (
        a[:, None] / jnp.sum(jnp.exp(log_K), axis=1, keepdims=True)
    )  # Normalized

    return phi, psi, plan


# Gradient using optax
@jax.jit
def gradient_ascent_opt_multimarginal(
    C,
    marginals,
    eps=1e-3,
    learning_rate=1e-3,
    max_iterations=100_000,
    tol=1e-4,
) -> jax.Array:
    N = marginals.shape[0]
    n = marginals.shape[1]

    shapes = [tuple(n if j == i else 1 for j in range(N)) for i in range(N)]
    potentials = jnp.zeros_like(marginals)
    optimizer = optax.sgd(learning_rate=learning_rate)
    opt_state = optimizer.init(potentials)

    @jax.jit
    def objective(potentials: jax.Array):
        """Computes the dual objective with logsumexp stabilization."""
        potentials_reshaped = [potentials[i].reshape(shapes[i]) for i in range(N)]
        potentials_sum = sum(potentials_reshaped)
        log_sub_entropy = (potentials_sum - C) / eps
        max_log_sub_entropy = jnp.max(log_sub_entropy, axis=0, keepdims=True)
        stable_sum = jnp.exp(max_log_sub_entropy) * jnp.sum(
            jnp.exp(log_sub_entropy - max_log_sub_entropy), axis=0
        )
        dual = potentials * marginals
        return jnp.sum(dual - eps * stable_sum)

    objective_gradient = jax.value_and_grad(objective)

    @jax.jit
    def step(state: tuple[int, jax.Array, optax.OptState, float, bool]):
        """Performs one gradient ascent step."""
        i, potentials, opt_state, prev_loss, _ = state
        loss, grad = objective_gradient(potentials)
        # minus gradient because we are performing gradient ascent
        updates, opt_state = optimizer.update(-grad, opt_state, potentials)
        potentials = optax.apply_updates(potentials, updates)
        # L-infinity norm
        max_change = jnp.max(jnp.abs(potentials - state[1]))
        has_converged = max_change < tol
        return i + 1, potentials, opt_state, loss, has_converged
    
    def cond_fn(state: tuple[int, jax.Array, optax.OptState, float, bool]):
        i, _, _, _, has_converged = state
        return jnp.logical_and(i < max_iterations, jnp.logical_not(has_converged))

    final_state = jax.lax.while_loop(
        cond_fn, step, (0, potentials, opt_state, jnp.inf, False)
    )

    steps, final_potentials, _, final_loss, has_converged = final_state
    return final_potentials, has_converged

def gradient_ascent_two_marginal(a, b, C, epsilon=1e-3):
    marginals = jnp.array([a, b])
    final_potentials = gradient_ascent_opt_multimarginal(C, marginals, eps=epsilon, learning_rate=1e-3).block_until_ready()
    return final_potentials[0][None, :].reshape(-1), final_potentials[1][:, None].reshape(-1)


def gradient_ascent(a, b, C, epsilon = 1e-3):
    marginals = jnp.array([a, b])
    
    final_potentials, converged = gradient_ascent_opt_multimarginal(C, marginals, eps=epsilon, learning_rate=1e-3)
    final_potentials.block_until_ready()
    converged.block_until_ready()

    P = jnp.exp(
        (final_potentials[0][None, :] + final_potentials[1][:, None] - C) / epsilon
    )

    return P, jnp.sum(C * P), converged



import numpy as np
import pandas as pd
# from IPython.display import display, HTML

def get_agg_table(df, metrics: list[str]):
        aggregation_rules = {f"{metric}_mean": (metric, 'mean') for metric in metrics} | \
                            {f"{metric}_std": (metric, lambda x: np.std(x, ddof=1)) for metric in metrics }

        
        df = df.groupby('dataset').agg(**aggregation_rules).reset_index()
        return df


def get_mean_comparison_table(df: dict[str, pd.DataFrame], field: str):
        comparison_df = pd.DataFrame()
        comparison_df['dataset'] = df.dataset.unique()

        dfs = [df[df.name == name] for name in df.name.unique()]
        agg_dfs = {df.name.iloc[0]: get_agg_table(df, [field]) for df in dfs}

        for name, df in agg_dfs.items():
                mean_column = f"{field}_mean"
                df = df[['dataset', mean_column]].rename(columns={mean_column: name})
                comparison_df = pd.merge(comparison_df, df, on='dataset', how='inner')
        return comparison_df

def get_std_comparison_table(df: pd.DataFrame, field: str):
        comparison_df = pd.DataFrame()
        comparison_df['dataset'] = df.dataset.unique()

        dfs = [df[df.name == name] for name in df.name.unique()]
        agg_dfs = {df.name.iloc[0]: get_agg_table(df, [field]) for df in dfs}

        for name, df in agg_dfs.items():
                std_column = f"{field}_std"
                df = df[['dataset', std_column]].rename(columns={std_column: name})
                comparison_df = pd.merge(comparison_df, df, on='dataset', how='inner')
        return comparison_df

def display_mean_and_std(agg_dfs, column: str):
        def highlight_top_two(s):
                is_first = s == s.min()
                is_second = s == s[s != s.min()].min()
                return [
                        'background-color: lightgreen' if f else
                        'background-color: lightblue' if s_ else
                        ''
                        for f, s_ in zip(is_first, is_second)
                ]
        mean_comparison = get_mean_comparison_table(agg_dfs, column)
        std_comparison = get_std_comparison_table(agg_dfs, column)

        mean_html = mean_comparison.style.apply(highlight_top_two, axis=1, subset=mean_comparison.columns[1:]).to_html()

        std_html = std_comparison.style.apply(highlight_top_two, axis=1, subset=std_comparison.columns[1:]).to_html()

        combined_html = f"""
        <h3 style="text-align:center;">{column} mean and std</h3>
        <div style="display: flex; justify-content: space-around;">
        <div>{mean_html}</div>
        <div>{std_html}</div>
        </div>
        """
        return display(HTML(combined_html))

def display_all_metrics(results: dict[str, pd.DataFrame], metrics_names: list[str]):
        agg_dfs = {solver_name: get_agg_table(result.df, metrics=metrics_names) for solver_name, result in results.items()}

        for metrics_name in metrics_names:
                display_mean_and_std(agg_dfs, metrics_name)


import random

import numpy as np
from scipy.optimize import linprog
from scipy.linalg import lstsq
from matplotlib import pyplot as plt


##############################################################################################
# Taken from https://github.com/aipyth/gencol_ot
# and modified to work with the current codebase.
##############################################################################################


def place_at_bin(arr, N):
    max_choice = int(N - arr.sum())
    if max_choice <= 0:
        max_choice = 1
    i = random.randint(0, max_choice)
    index = random.choice(np.where(arr == 0)[0])
    arr[index] = i


def generate_random_cols(N, matrix_size):
    A = np.zeros(matrix_size)
    i = 0
    while i < matrix_size[1]:
        while A[:, i].sum() != N:
            place_at_bin(A[:, i], N)
        # check for any repetitions
        for j in range(i):
            if np.all(A[:, j] == A[:, i]):
                A[:, i] = 0
                i -= 1
        i += 1
    return A / N


def initialize_AI(N, grid_size, beta=5):
    A = np.eye(grid_size)
    random_cols = generate_random_cols(N, (grid_size, beta - 1))
    return np.hstack((A, random_cols))


def solve_rmp(AI, cI, marginal):
    result = linprog(cI, A_eq=AI, b_eq=marginal, bounds=(0, None))
    return result.x, result.fun


def solve_rmp_appprox(AI, cI, marginal):
    alpha, residuals, rank, s = lstsq(AI, marginal)
    return alpha


def solve_dual(AI, cI, marginal):
    c_dual = -marginal  # because linprog does minimization
    result = linprog(c_dual, A_ub=AI.T, b_ub=cI,
                     bounds=[(None, 0) for _ in range(AI.shape[0])],
                     method='highs')
    return result.x


def solve_dual_of_rmp(AI, cI, marginal):
    # Negate lambda_star to convert the maximization problem to a minimization problem
    c_dual = -marginal

    # Transpose AI because linprog works with constraints in the form A_ub @ x <= b_ub
    A_ub = AI.T

    # Bounds for y, each y_i should be unrestricted in negative direction and 0 in positive
    # (as `None` indicates no bound and linprog treats variables as non-negative by default)
    bounds = [(None, 0) for _ in range(A_ub.shape[1])]

    # Solve the problem using linprog
    res = linprog(c_dual, A_ub=A_ub, b_ub=cI, bounds=bounds, method='highs')

    # Check if the optimization was successful
    if res.success:
        # Return the optimal value of the original maximization problem by negating res.fun
        optimal_value = -res.fun
        y = res.x
        # print("Optimal dual variables y:", y)
        # print("Maximal value of y^T lambda_star:", optimal_value)
    else:
        print("Optimization failed:", res.message)
        y = None
        optimal_value = None

    return y


def mutate_parent(parent, grid_size):
    child = parent.copy()
    change_index = random.choice(np.where(parent > 0)[0])
    direction = random.choice([-1, 1])
    destination = max(min(grid_size-1, change_index + direction), 0)
    child[destination] += parent[change_index]
    child[change_index] -= parent[change_index]
    child /= child.sum()
    return child


def compute_cost(lambd, N, cost_matrix: np.ndarray):
    return (N**2/2 * lambd.T @ cost_matrix @ lambd -
            N/2 * cost_matrix.diagonal().T @ lambd)


def genetic_column_generation(
        N,
        l,
        beta,
        pair_potential,
        coordinates_of_sites,
        marginal,
        maxiter,
        maxsamples
):
    a, b = coordinates_of_sites
    x = y = np.linspace(a, b, l)
    X, Y = np.meshgrid(x, y, indexing='ij')
    cost_matrix = pair_potential(X, Y)

    # AI = np.eye(l)
    # AI = np.random.randn(l, l)
    # AI /= AI.sum(axis=0)
    AI = initialize_AI(N, l)

    cI = np.empty(AI.shape[1])
    for j in range(AI.shape[1]):
        cI[j] = compute_cost(AI[:, j], N, cost_matrix)
    samples = 0
    iter = 0
    gain = -1

    cost_history = np.empty((maxiter, 1))
    dual_value_history = np.empty((maxiter, l))

    for i in range(maxiter):
        alpha_I, cost = solve_rmp(AI, cI, marginal)
        cost_history[i] = cost
        y_star = solve_dual(AI, cI, marginal)
        dual_value_history[i, :] = y_star

        while gain <= 0 and samples <= maxsamples:
            # Select a random active column of AI
            parent_index = random.choice(np.where(alpha_I > 0)[0])
            parent = AI[:, parent_index]
            child = mutate_parent(parent, l)
            c_child = compute_cost(child, N, cost_matrix)

            # Calculate gain from adding the child column
            gain = np.dot(child.T, y_star) - c_child

            samples += 1

        # Update AI and cI with the new child column if there's a positive gain
        if gain > 0:
            AI = np.hstack((AI, child[:, np.newaxis]))
            # cI = np.hstack((cI, c_child))
            cI = np.append(cI, c_child)
            if AI.shape[1] > beta * l:
                # Clear the oldest inactive columns
                inactive_indices = np.where(alpha_I == 0)[0]
                AI = np.delete(AI, inactive_indices[:l], axis=1)
                cI = np.delete(cI, inactive_indices[:l])

        iter += 1


    # Return the final set of columns and configuration
    alpha_I, cost = solve_rmp(AI, cI, marginal)
    cost_history[i] = cost
    y_star = solve_dual(AI, cI, marginal)
    dual_value_history[i, :] = y_star
    return AI, alpha_I, cost_history, dual_value_history


def col_gen(a, b, C, N=6, maxiter=100, maxsamples=1000, beta=5):
    """
    Column generation algorithm for optimal transport.
    
    Args:
        a (np.ndarray): Source distribution.
        b (np.ndarray): Target distribution.
        C (np.ndarray): Cost matrix.
    
    Returns:
        tuple: (T, cost) where T is the optimal transport matrix and cost is the optimal cost.
    """
    grid_size = len(a)

    
    def pair_potential(x, y, eps=0.1):
        return C
    
    ai, alpha, cost_history, dual_value_history = genetic_column_generation(
        N, grid_size, beta, pair_potential, (0, 1), a, maxiter, maxsamples)
    
    T = np.zeros((len(a), len(b)))
    

    for j in range(ai.shape[1]):
        col_contribution = np.outer(ai[:, j], b) * alpha[j]
        T += col_contribution
    

    T = T / T.sum() * a.sum()
    
    cost = np.sum(T * C)
    
    return T, cost


if __name__ == "__main__":
    N = 6  # Number of marginals
    grid_size = 50  # Number of sites
    beta = 5  # Hyperparameter for controlling the maximum columns
    maxiter = 500  # Maximum number of iterations
    maxsamples = 5000  # Maximum number of samples for mutations
    # maxsamples = 100  # Maximum number of samples for mutations

    grid_points = np.arange(1, grid_size+1)

    marginal = 0.2 + np.power(np.sin(np.pi * grid_points / (grid_size+1)), 2)
    marginal /= marginal.sum()

    def pair_potential(x, y, eps=0.1):
        "Use regulatized Coulomb interaction"
        return 1 / np.sqrt(eps**2 + np.power(x - y, 2))

    print("1-DIMENSIONAL SETTING.")
    print(f"Number of marginals {N=}")
    print(f"Number of grid points {grid_size=}")
    print(f"Hyperparameter for controlling the maximum columns {beta=}")
    print(f"Maximum number of iterations {maxiter=}")
    print(f"First 6 gridpoints {grid_points[:6]}")
    print(f"Last 6 gridpoints {grid_points[-6:]}")
    print(f"Marginal {marginal[:5]}")

    print("=" * 100)

    ai, alpha, cost_history, dual_value_history = genetic_column_generation(
        N, grid_size, beta, pair_potential, (0, 1), marginal, maxiter, maxsamples)

    print(f"{ai.shape=}")
    print(f"{alpha.shape=}")

    gamma = ai @ alpha

    plt.scatter(grid_points, gamma)
    plt.show()


import time
import numpy as np
from uot.algorithms.sinkhorn import sinkhorn
from memory_profiler import memory_usage, profile
from uot.core.experiment import OTProblem, Experiment

def time_precision_experiment_func(ot_problem: OTProblem, solver: callable = sinkhorn):
    a, b, C = ot_problem.to_jax_arrays()
    start_time = time.perf_counter()
    output_T, output_dist, converged = solver(a, b, C)
    end_time = time.perf_counter()

    exact_T, exact_dist = ot_problem.exact_map, ot_problem.exact_cost
    precision = np.abs(output_dist - exact_dist) / exact_dist
    coupling_precision = np.sum(np.abs(output_T - exact_T)) / np.prod(output_T.shape)

    return {'time': (end_time - start_time) * 1000, 'cost_rerr': precision, 'coupling_avg_err': coupling_precision, 'converged': converged}


def time_experiment_func(ot_problem: OTProblem, solver: callable = sinkhorn):
    a, b, C = ot_problem.to_jax_arrays()
    start_time = time.perf_counter()
    _, _ = solver(a, b, C)
    end_time = time.perf_counter()

    return {'time': (end_time - start_time) * 1000}


# Requires if __name__ == '__main__' to work properly
def memory_experiment(ot_problem: OTProblem, solver: callable = sinkhorn):
    a, b, C = ot_problem.to_jax_arrays()

    mem_usage = memory_usage((solver, (a, b, C)), interval=1e-3)

    return {'memory': max(mem_usage)}


def memory_profiler_out(ot_problem: OTProblem, solver: callable = sinkhorn):
    a, b, C = ot_problem.to_jax_arrays()

    profiled_solver = profile(solver)
    profiled_solver(a, b, C)


time_precision_experiment = Experiment("Measure time and precision", run_function=time_precision_experiment_func)

time_experiment = Experiment("Measure time", run_function=time_experiment_func)





import jax
jax.config.update("jax_enable_x64", True)

import time
from uot.algorithms.lp import pot_lp
from uot.algorithms.sinkhorn import jax_sinkhorn
from uot.core.experiment import generate_data_problems, get_problemset

problem = get_problemset((3, "3dmesh", 1024))[0]
print(problem)

# N = 10
# a, b, C = problem.to_jax_arrays()

# total = 0
# for i in range(N):
#     print(i)
#     start = time.perf_counter()
#     _, cost = pot_lp(a, b, C)
#     end = time.perf_counter()
#     total += end - start

# print("LP:",  1000 * total / N)

# total = 0
# for i in range(N):
#     print(i)
#     start = time.perf_counter()
#     _, cost = jax_sinkhorn(a, b, C)
#     end = time.perf_counter()
#     total += end - start
# print("JAX:",  1000 * total / N)




import zipfile
import requests
import open3d as o3d
import numpy as np
import pandas as pd
import scipy.stats as stats
import os

import matplotlib.pyplot as plt
from matplotlib import cm
from itertools import product

def compare_measure_kwargs(source: dict, target: dict):
    if source.keys() != target.keys():
        return False  

    for key in source:
        if isinstance(source[key], np.ndarray) and isinstance(target[key], np.ndarray):
            if not np.array_equal(source[key], target[key]):
                return False  
        elif source[key] != target[key]:
            return False

    return True  

class Measure:

    def __init__(self, name, support: list[np.ndarray], distribution: np.ndarray, kwargs=None):
        self.name = name
        self.support = support
        self.distribution = distribution
        self.kwargs = kwargs
    
    def to_histogram(self):
        """
        Converts the support and distribution data into histogram format.
        """
        support = self.get_flat_support()
        distribution = self.distribution.ravel()
        return support, distribution
    
    def get_flat_support(self):
        return np.vstack([coordinate.ravel() for coordinate in self.support]).T
        
    def __str__(self):
        """
        Returns a string representation of the Measure object, displaying its name and kwargs.
        """
        kwargs_str = ", ".join(f"{key}={value}" for key, value in (self.kwargs or {}).items())
        return f"Measure(name='{self.name}', kwargs={{ {kwargs_str} }})".replace('\n', '')

    def __repr__(self):
        return str(self) + '\n'
    
    def __hash__(self):
        points, distribution = self.to_histogram()
        return hash(self.name) & hash(tuple(points.flatten())) & \
               hash(tuple(distribution.flatten())) & hash(tuple(self.kwargs.keys())) & \
               hash(str(self.kwargs.values())) 
    
    def __eq__(self, other):
        if not isinstance(other, Measure):
            return False
        return self.name == other.name and \
               np.array_equal(self.support, other.support) and \
               np.array_equal(self.distribution, other.distribution) and \
               compare_measure_kwargs(self.kwargs, other.kwargs)

    def plot(self):
        if len(self.support) == 1:
            plt.figure(figsize=(8, 4))
            plt.plot(self.support[0], self.distribution, label=self.name)
            plt.xlabel("Support")
            plt.ylabel("Distribution")
            plt.title(f"1D Distribution: {self.name}")
            plt.legend()
            plt.grid(True)
            plt.show()
        elif len(self.support) == 2:
            print(self.distribution)
            x, y = self.support
            plt.figure(figsize=(8, 6))
            plt.imshow(self.distribution, extent=[x.min(), x.max(), y.min(), y.max()], origin='lower', cmap='viridis')
            plt.colorbar(label="Density")
            plt.title("2D Gaussian Distribution")
            plt.xlabel("X-axis")
            plt.ylabel("Y-axis")
            plt.grid(False)
            plt.show()
        elif len(self.support) == 3:
            points, distribution = self.to_histogram()
            normalized_distribution =  (distribution - np.min(distribution)) / (np.max(distribution) - np.min(distribution))
            colormap = cm.get_cmap('viridis')
            colors = colormap(normalized_distribution)[:, :3]  
            point_cloud = o3d.geometry.PointCloud()
            point_cloud.points = o3d.utility.Vector3dVector(points)
            point_cloud.colors = o3d.utility.Vector3dVector(colors)
            o3d.visualization.draw_geometries([point_cloud])
        

def generate_gaussian_pdf(x, mean=0, std=1):
    pdf = stats.norm.pdf(x, loc=mean, scale=std)
    pdf_sum = pdf.sum()
    if pdf_sum > 1e-10:
        return pdf / pdf_sum
    else:
        return np.ones_like(pdf) / len(pdf)


def generate_2d_gaussian_pdf(x, y, mean=(0, 0), cov=((1, 0), (0, 1))):
    """
    Generates a 2D Gaussian PDF on a mesh grid.

    Args:
        x (np.array): X-coordinates of the mesh grid.
        y (np.array): Y-coordinates of the mesh grid.
        mean (tuple): Mean of the Gaussian distribution (default is (0, 0)).
        cov (tuple): Covariance matrix of the Gaussian distribution 
                     (default is identity matrix).

    Returns:
        np.array: 2D Gaussian PDF values on the mesh grid.

    Usage:
        x = np.linspace(-2, 2, 2)
        y = np.linspace(-2, 2, 2)

        x, y = np.meshgrid(x, y)

        gaussian = generate_2d_gaussian_pdf(x, y)
    """
    pos = np.stack((x, y), axis=-1)
    rv = stats.multivariate_normal(mean=mean, cov=cov)
    pdf = rv.pdf(pos)
    return pdf / pdf.sum()


def generate_3d_gaussian_pdf(x, y, z, mean=(0, 0, 0), cov=((1, 0, 0), (0, 1, 0), (0, 0, 1))):
    """
    Generates a 3D Gaussian PDF on a mesh grid.

    Args:
        x (np.array): X-coordinates of the mesh grid.
        y (np.array): Y-coordinates of the mesh grid.
        z (np.array): Z-coordinates of the mesh grid.
        mean (tuple): Mean of the Gaussian distribution (default is (0, 0, 0)).
        cov (tuple): Covariance matrix of the Gaussian distribution 
                     (default is identity matrix).

    Returns:
        np.array: 3D Gaussian PDF values on the mesh grid.

    Usage:
        x = np.linspace(-2, 2, 2)
        y = np.linspace(-2, 2, 2)
        z = np.linspace(-2, 2, 2)

        x, y, z = np.meshgrid(x, y, z)

        gaussian = generate_3d_gaussian_pdf(x, y, z)
    """
    pos = np.stack((x, y, z), axis=-1)
    rv = stats.multivariate_normal(mean=mean, cov=cov)
    pdf = rv.pdf(pos)
    return pdf / pdf.sum()


def generate_gamma_pdf(x, shape=1, scale=1):
    pdf = stats.gamma.pdf(x, a=shape, scale=scale)
    return pdf / pdf.sum()
    

def generate_beta_pdf(x, alpha=1, beta=1):
    pdf = stats.beta.pdf(x, a=alpha, b=beta)
    pdf = np.nan_to_num(pdf, nan=0.0, posinf=0, neginf=0)
    return pdf / pdf.sum()

def generate_uniform_pdf(x, lower=0, width=1):
    pdf = stats.uniform.pdf(x, loc=lower, scale=width)
    return pdf / pdf.sum()

def generate_cauchy_pdf(x, loc=0, scale=1):
    pdf = stats.cauchy.pdf(x, loc=loc, scale=scale)
    return pdf / pdf.sum()

def generate_normalized_white_noise(x, mean=0, std=1):
    size = x.shape
    noise = np.random.normal(loc=mean, scale=std, size=size)
    return (noise - np.mean(noise)) / np.std(noise)

def get_ts_dataset() -> list[np.array]:
    """
    Reads a time series dataset from a CSV file, processes it by region, and returns 
    a list of normalized time series data.

    Returns:
        list: A list of numpy arrays, where each array represents the normalized 
              time series data for a specific region.
    """
    df = pd.read_csv('datasets/ts_dataset.csv')
    regions = df.region.unique()
    time_series = []
    for region in regions:
        ts = df[df.region == region].sort_values(by="year")['tincidence']
        ts /= ts.sum()
        time_series.append(ts.to_numpy())
    return time_series
 

def generate_3d_gaussian_pdf(x, y, z, mean=(0, 0, 0), cov=((1, 0, 0), (0, 1, 0), (0, 0, 1))):
    """
    Generates a 3D Gaussian PDF on a mesh grid.

    Args:
        x (np.array): X-coordinates of the mesh grid.
        y (np.array): Y-coordinates of the mesh grid.
        z (np.array): Z-coordinates of the mesh grid.
        mean (tuple): Mean of the Gaussian distribution (default is (0, 0, 0)).
        cov (tuple): Covariance matrix of the Gaussian distribution 
                     (default is identity matrix).

    Returns:
        np.array: 3D Gaussian PDF values on the mesh grid.

    Usage:
        x = np.linspace(-2, 2, 2)
        y = np.linspace(-2, 2, 2)
        z = np.linspace(-2, 2, 2)

        x, y, z = np.meshgrid(x, y, z)

        gaussian = generate_3d_gaussian_pdf(x, y, z)
    """
    pos = np.stack((x, y, z), axis=-1)
    rv = stats.multivariate_normal(mean=mean, cov=cov)
    return rv.pdf(pos)

def generate_normalized_white_noise_2d(x, y, mean=0, std=1):
    """
    Generates normalized white noise on a 2D mesh grid.

    Args:
        x (np.array): X-coordinates of the mesh grid.
        y (np.array): Y-coordinates of the mesh grid.
        mean (float): Mean of the white noise distribution (default is 0).
        std (float): Standard deviation of the white noise distribution (default is 1).

    Returns:
        np.array: Normalized white noise values on the 2D mesh grid.
    """
    shape = x.shape
    mean = mean[0]
    std = std[0]
    noise = np.random.normal(loc=mean, scale=std, size=shape)
    return (noise - np.mean(noise)) / np.std(noise)


def generate_normalized_white_noise_3d(x, y, z, mean=0, std=1):
    """
    Generates normalized white noise on a 3D mesh grid.

    Args:
        x (np.array): X-coordinates of the mesh grid.
        y (np.array): Y-coordinates of the mesh grid.
        z (np.array): Z-coordinates of the mesh grid.
        mean (float): Mean of the white noise distribution (default is 0).
        std (float): Standard deviation of the white noise distribution (default is 1).

    Returns:
        np.array: Normalized white noise values on the 3D mesh grid.
    """
    shape = x.shape
    mean = mean[0]
    std = std[0]
    noise = np.random.normal(loc=mean, scale=std, size=shape)
    return (noise - np.mean(noise)) / np.std(noise)

def generate_coefficients(dim: int, distributions: dict[str, int]):
    if dim not in [1, 2, 3]:
        raise ValueError("dim must be 1, 2, or 3.")

    basic_ranges = {
        'mean_range': (-5, 5),
        'std_range': (0.5, 4.0),
        'shape_range': (1, 5),
        'scale_range': (0.2, 2.0),
        'loc_range': (-5, 5),
        'alpha_range': (0.5, 8),
        'beta_range': (0.5, 8),
        'width_range': (2, 8),
        'lower_range': (-8, 5),
    }

    distribution_parameters = {
        'gaussian': ('mean', 'std'),
        'gamma': ('shape', 'scale'),
        'beta': ('alpha', 'beta'),
        'uniform': ('lower', 'width'),
        'cauchy': ('loc', 'scale'),
        'white-noise': ('mean', 'std'),
    }

    results = {}

    for distribution in distributions:
        if distribution not in distribution_parameters:
            raise ValueError(f"Unsupported distribution: {distribution}")

        num_to_generate = distributions[distribution]
        param_names = distribution_parameters[distribution]
        param_ranges = []

        for param in param_names:
            if f"{param}_range" not in basic_ranges:
                raise ValueError(f"Missing range for {param}.")
            
            num_points = max(10, num_to_generate * 2)
            values = np.linspace(*basic_ranges[f"{param}_range"], num_points)
            
            if dim == 1:
                param_ranges.append(values)
            else:
                param_ranges.append(list(product(values, repeat=dim)))

        if distribution == 'gaussian' and dim > 1:
            mean_values = np.linspace(*basic_ranges['mean_range'], max(10, num_to_generate))
            mean_choices = list(product(mean_values, repeat=dim))
            
            step = max(1, len(mean_choices) // num_to_generate)
            selected_means = mean_choices[::step][:num_to_generate]
            
            result = []
            
            diag_values = np.linspace(0.5, 3.0, 10)
            offdiag_values = np.linspace(-0.3, 0.3, 7)
            
            for i, mean in enumerate(selected_means):
                diag_indices = [(i + j) % len(diag_values) for j in range(dim)]
                offdiag_idx = i % len(offdiag_values)
                
                diag = np.array([diag_values[idx] for idx in diag_indices])
                cov = np.diag(diag)
                
                indices = np.triu_indices(dim, k=1)
                for idx in range(len(indices[0])):
                    row, col = indices[0][idx], indices[1][idx]
                    val = offdiag_values[offdiag_idx]
                    cov[row, col] = cov[col, row] = val
                
                min_eig = np.min(np.linalg.eigvalsh(cov))
                if min_eig <= 0:
                    cov += np.eye(dim) * (abs(min_eig) + 1e-6)
                
                result.append((tuple(np.round(mean, 2)), np.round(cov, 2)))
            
            results[distribution] = result
        else:
            all_combinations = list(product(*param_ranges))
            
            step = max(1, len(all_combinations) // num_to_generate)
            selected_combinations = all_combinations[::step][:num_to_generate]

            rounded_combinations = [
                tuple(np.round(combination, 2)) for combination in selected_combinations
            ]
            results[distribution] = rounded_combinations

    return results


def generate_grid(dim: int, grid_size: int, start: int = -10, end: int = 10):
    """
    Generates a mesh grid for the specified dimensionality and size.

    Args:
        dim (int): Dimensionality of the grid (1, 2, or 3).
        grid_size (int): Number of points along each dimension.
        start (int): Starting point of the grid (default is -5).
        end (int): Ending point of the grid (default is 5).

    Returns:
        list[np.ndarray]: A list of mesh grid arrays for the specified dimensions.
    """
    if dim not in [1, 2, 3]:
        raise ValueError("dim must be 1, 2, or 3.")

    if dim == 1:
        return [np.linspace(start, end, grid_size)]

    elif dim == 2:
        x = np.linspace(start, end, grid_size)
        y = np.linspace(start, end, grid_size)
        return np.meshgrid(x, y)

    elif dim == 3:
        x = np.linspace(start, end, grid_size)
        y = np.linspace(start, end, grid_size)
        z = np.linspace(start, end, grid_size)
        return np.meshgrid(x, y, z)


def get_grids(dim: list[int], grid_sizes: list[int], start: int = -10, end: int = 10):
    '''
    Generates a list of mesh grids for the specified dimensions and sizes.
    '''
    grids = {}
    for grid_size in grid_sizes:
        grids[f"{'x'.join([str(grid_size)] * dim)} {dim}D"] = generate_grid(dim, grid_size, start, end)
    
    return grids

def generate_measures(dim: int, coefficients: dict[str, list[tuple]], grids: list[np.ndarray]):
    """
    Generates measures based on the provided coefficients and grid.

    Args:
        dim (int): Dimensionality of the measures (1, 2, or 3).
        coefficients (dict): Coefficients for the distributions.
        grid (list[np.ndarray]): List of mesh grid arrays for the dimensions.

    Returns:
        list[Measure]: A list of Measure objects with randomly generated characteristics.
    """
    if dim not in [1, 2, 3]:
        raise ValueError("dim must be 1, 2, or 3.")
    
    distribution_map = {
        'gaussian': {
            1: generate_gaussian_pdf,
            2: generate_2d_gaussian_pdf,
            3: generate_3d_gaussian_pdf
        },
        'gamma': {
            1: generate_gamma_pdf,
        },
        'beta': {
            1: generate_beta_pdf,
        },
        'uniform': {
            1: generate_uniform_pdf,
        },
        'cauchy': {
            1: generate_cauchy_pdf,
        },
        'white-noise': {
            1: generate_normalized_white_noise,
            2: generate_normalized_white_noise_2d,
            3: generate_normalized_white_noise_3d
        }
    }

    output = {}

    for name, grid in grids.items():

        entry = f"{name} {'|'.join(sorted(param for param in coefficients))}"
        output[entry] = []

        for distribution in coefficients:

            func = distribution_map[distribution].get(dim)
            if func is None:
                raise ValueError(f"Unsupported distribution for {dim}D: {distribution}.")

            for params in coefficients[distribution]:
                pdf = func(*grid, *params)
                mes = Measure(name=f"{name} {distribution}", support=grid, distribution=pdf, kwargs={distribution: params})
                output[entry].append(mes)
        
    return output

def download_dataset():
    URL = "https://drive.usercontent.google.com/download?id=1h2LA05z19P1BWUH5v2ph0gvXKS_wxS8W&export=download&confirm=t&uuid=8ec0d845-fa17-4a8e-8211-ac9214564b85&at=APcmpoxvi8f0AvQwJflzKLsBtg7_:1746613211647"
    FILENAME = "dataset.zip"

    with requests.get(URL, stream=True) as r:
        r.raise_for_status()  # raise an exception for bad status codes
        with open(FILENAME, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive chunks
                    f.write(chunk)
    
    with zipfile.ZipFile("dataset.zip", 'r') as zip_ref:
        zip_ref.extractall(".")

    os.remove(FILENAME) 

if __name__ == "__main__":
    coeffs = generate_coefficients(1, {'gaussian': 1, 'gamma': 1, 'beta': 1, 'uniform': 1, 'cauchy': 1, 'white-noise': 1})
    grids = get_grids(1, [64])
    measures = generate_measures(1, coeffs, grids)
    for key, value in measures.items():
        for measure in value:
            print(measure)
            measure.plot()


import os
os.environ["JAX_PLATFORM_NAME"] = "cpu"

import jax
jax.config.update("jax_enable_x64", True)

import ot
import numpy as np
from tqdm import tqdm

from sklearn.datasets import load_digits
from sklearn.preprocessing import normalize

from algorithms.sinkhorn import jax_sinkhorn, pot_sinkhorn
from algorithms.gradient_ascent import gradient_ascent
from algorithms.lbfgs import dual_lbfgs
from algorithms.lp import pot_lp

import multiprocessing as mp
from functools import partial
import argparse
import matplotlib.pyplot as plt

digits = load_digits()
X, _ = digits.data, digits.target
X = X / X.sum(axis=1).reshape(X.shape[0],1) + 1e-12
X = X[:100]


solvers = {
    'sinkhorn': partial(jax_sinkhorn, epsilon=1e-1),
    'grad-ascent': partial(gradient_ascent, epsilon=1e-1),
    'dual-lbfs': dual_lbfgs,
    'pot-sinkhorn': partial(pot_sinkhorn, epsilon=1e-1),
    'lp': pot_lp
}

row, col = np.arange(8), np.arange(8)
row, col = np.meshgrid(row, col)
points = np.vstack([coordinate.ravel() for coordinate in [row, col]]).T
C = ot.dist(points, points).astype('float64')
C /= C.max()

pairwise_distances = {}

def compute_distance_pair(coordinates, X, C, solver_fn):
    i, j = coordinates
    dist = solver_fn(X[i], X[j], C)[1]
    return (i, j, dist)

def compute_distances_parallel(X, C, solver_name, solver_fn, num_processes):
    n = X.shape[0]
    args = [(i, j) for i in range(n) for j in range(i, n) if i < j]
    
    with mp.Pool(num_processes) as pool:
        worker = partial(compute_distance_pair, X=X, C=C, solver_fn=solver_fn)
        results = list(tqdm(pool.imap_unordered(worker, args),
                            total=len(args),
                            desc=f"Computing distances with {solver_name}"))

    dist_matrix = np.zeros((n, n))
    for i, j, dist in results:
        dist_matrix[i, j] = dist_matrix[j, i] = dist

    np.savetxt(f"classification/{solver_name}_pairwise_distances.csv", dist_matrix, delimiter=",")

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Compute pairwise distances using a specified solver.")
    parser.add_argument("--solver", type=str, choices=solvers.keys(), required=True, help="Solver name to use.")
    parser.add_argument("--num-processes", type=int, default=4, help="Number of processes to use in the pool.")
    args = parser.parse_args()

    solver_name = args.solver
    solver_fn = solvers[solver_name]
    num_processes = args.num_processes

    compute_distances_parallel(X, C, solver_name, solver_fn, num_processes)


import jax
jax.config.update("jax_enable_x64", True)

import os
import itertools
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from ot.utils import dist

from uot.algorithms.sinkhorn import sinkhorn
from uot.algorithms.gradient_ascent import gradient_ascent_two_marginal
from uot.algorithms.lbfgs import dual_lbfs_potentials
from uot.algorithms.lp import pot_lp
from uot.algorithms.col_gen import col_gen


rng = np.random.RandomState(42)

def im2mat(img):
    """Converts and image to matrix (one pixel per line)"""
    return img.reshape((img.shape[0] * img.shape[1], img.shape[2]))

def mat2im(X, shape):
    """Converts back a matrix to an image"""
    return X.reshape(shape)

def minmax(img):
    return np.clip(img, 0, 1)

def get_samples(X1, X2, nb=500):
    idx1 = rng.randint(X1.shape[0], size=(nb,))
    idx2 = rng.randint(X2.shape[0], size=(nb,))

    Xs = X1[idx1, :]
    Xt = X2[idx2, :]
    return Xs, Xt

def read_image(image_path: str):
    image = plt.imread(image_path).astype(np.float64) / 256
    return image[
            (image.shape[0] - 640) // 2 : (image.shape[0] + 640) // 2,
            (image.shape[1] - 640) // 2 : (image.shape[1] + 640) // 2
        ]

def show_images(result):
    _, axes = plt.subplots(len(images_pairs), len(algorithms) + 2, figsize=(15, 6))

    for image_ind in range(len(results)):

        axes[image_ind, 0].imshow(images_pairs[image_ind][0])
        axes[image_ind, 0].axis("off")
        if image_ind == 0:
            axes[image_ind, 0].set_title("Source")

        for i, (name, img) in enumerate(results[image_ind].items()):
            axes[image_ind, i + 1].imshow(img)
            axes[image_ind, i + 1].axis("off")
            if image_ind == 0:
                axes[image_ind, i + 1].set_title(name)

        axes[image_ind, -1].imshow(images_pairs[image_ind][1])
        axes[image_ind, -1].axis("off")

        if image_ind == 0:
            axes[image_ind, -1].set_title("Target")

    plt.tight_layout()
    plt.show()

def export_images(image_pairs, results: dict, save_path: str):
    for (source_name, source_image), (target_name, target_image) in image_pairs:
        _, axes = plt.subplots(1, len(algorithms) + 2, figsize=(15, 5))

        axes[0].imshow(source_image)
        axes[0].axis("off")

        for i, (name, img) in enumerate(results[(source_name, target_name)].items()):
            axes[i + 1].imshow(img)
            axes[i + 1].set_title(name)
            axes[i + 1].axis("off")

        axes[-1].imshow(target_image)
        axes[-1].axis("off")

        plt.tight_layout()
        plt.savefig(os.path.join(save_path, f"{source_name}-{target_name}.png"))


def get_cost_matrix(X: np.ndarray, Y: np.ndarray):
    return np.linalg.norm((X[:,None] - Y), axis=-1)


def entropic_transport(v: np.ndarray, C: np.ndarray, source_image: np.ndarray,
                       target_sample: np.ndarray, batch_size=128, epsilon=1e-2) -> np.ndarray:
    indices = np.arange(source_image.shape[0])
    batch_indices = [indices[i: i + batch_size] for i in range(0, len(indices), batch_size)]

    transported = []

    for i, batch_index in enumerate(batch_indices):
        C = get_cost_matrix(source_image[batch_index], target_sample)
        K = np.exp(-C/epsilon + v)
        transported_X = np.dot(K, target_sample) / np.sum(K, axis=1)[:, None]
        transported.append(transported_X)
    return np.concatenate(transported, axis=0)


def emd_transport(coupling: np.ndarray, source_image: np.ndarray, source_sample: np.ndarray,
                  target_sample: np.ndarray, batch_size=128):
    indices = np.arange(source_image.shape[0])
    batch_indices = [
        indices[i : i + batch_size]
        for i in range(0, len(indices), batch_size)
    ]

    transp_Xs = []
    for batch_index in batch_indices:
        D0 = dist(source_image[batch_index], source_sample)
        idx = np.argmin(D0, axis=1)

        transp = coupling / np.sum(coupling, axis=1)[:, None]
        transp = np.nan_to_num(transp, nan=0, posinf=0, neginf=0)
        transp_Xs_ = np.dot(transp, target_sample)

        transp_Xs_ = transp_Xs_[idx, :] + source_image[batch_index] - source_sample[idx, :]

        transp_Xs.append(transp_Xs_)

    transp_Xs = np.concatenate(transp_Xs, axis=0)

    return transp_Xs


SAMPLE_SIZE = 128
BATCH_SIZE = 20000
IMAGES_DIR = os.path.join("datasets", "color_images")
EPSILON = 1e-1
images = [(image_file, read_image(os.path.join(IMAGES_DIR, image_file)))
          for image_file in os.listdir(IMAGES_DIR)]

ENTROPIC_TRANSPORT = 'entropic'
EMD_TRANSPORT = 'emd'

algorithms = {
    'sinkhorn': (sinkhorn, ENTROPIC_TRANSPORT),
    'grad-ascent': (gradient_ascent_two_marginal, ENTROPIC_TRANSPORT),
    'dual-lbfs': (dual_lbfs_potentials, ENTROPIC_TRANSPORT),
    'lp': (pot_lp, EMD_TRANSPORT),
    'col_gen': (col_gen, EMD_TRANSPORT)
}

images_pairs = list(itertools.permutations(images, 2))

results = {}


with tqdm(total=len(images_pairs) * len(algorithms), desc="Running color transfer") as pbar:
    for source_image, target_image in images_pairs:
        source_name, source_image = source_image
        target_name, target_image = target_image

        results[(source_name, target_name)] = {}

        shape = source_image.shape
        source_image, target_image = im2mat(source_image), im2mat(target_image)
        source_sample, target_sample = get_samples(source_image, target_image, SAMPLE_SIZE)
        source_measure, target_measure = np.ones(SAMPLE_SIZE) / SAMPLE_SIZE, np.ones(SAMPLE_SIZE) / SAMPLE_SIZE
        C = get_cost_matrix(source_sample, target_sample)

        for solver_name, (solver, transport_name) in algorithms.items():
            pbar.update(1)

            if transport_name == ENTROPIC_TRANSPORT:
                _, v = solver(source_measure, target_measure, C, epsilon=EPSILON)
                transported_image = entropic_transport(v, C, source_image, target_sample, batch_size=BATCH_SIZE, epsilon=EPSILON)
            elif transport_name == EMD_TRANSPORT:
                P, _ = solver(source_measure, target_measure, C)
                transported_image = emd_transport(coupling=P, source_image=source_image,
                                                  source_sample=source_sample, target_sample=target_sample,
                                                  batch_size=BATCH_SIZE)
            
            transported_image = minmax(mat2im(transported_image, shape))
            results[(source_name, target_name)][solver_name] = transported_image


export_images(images_pairs, results, "color_transfer_results/")


import ot
import numpy as np

def pot_lp(a, b, C):
    a = np.asarray(a)
    b = np.asarray(b)
    C = np.asarray(C)
    T, log = ot.emd(a, b, C, log=True, numItermax=10000000)
    if log['warning'] is not None:
        print(log['warning'])
    converged = log['warning'] is None 
    return T, np.sum(T * C), converged


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from IPython.display import display, HTML
from uot.analysis import get_agg_table, get_mean_comparison_table, get_std_comparison_table, display_mean_and_std, display_all_metrics

plt.style.use('ggplot')
sns.set_palette("colorblind")

results_df = pd.read_csv('result_2025-05-01_02-52-55.csv')

print(f"Shape of results: {results_df.shape}")
print(f"Algorithms: {results_df['name'].unique()}")
print(f"Datasets: {results_df['dataset'].unique()}")

results_df.head()

metrics_to_analyze = ['time', 'cost_rerr', 'coupling_avg_err']

one_dim_datasets = [ds for ds in results_df['dataset'].unique() if '1D' in ds]
two_dim_datasets = [ds for ds in results_df['dataset'].unique() if '2D' in ds or 'x32' in ds]
mesh_datasets = [ds for ds in results_df['dataset'].unique() if 'Mesh' in ds]

dataset_groups = {
    '1D Datasets': one_dim_datasets,
    '2D Datasets': two_dim_datasets,
    '3D Mesh Datasets': mesh_datasets
}

algorithms = results_df['name'].unique()

import os
import matplotlib.pyplot as plt
import seaborn as sns

def plot_runtime_distributions(datasets, results_df, algorithms, save_path):
    os.makedirs(save_path, exist_ok=True)
    
    for dataset in datasets:
        dataset_data = results_df[results_df['dataset'] == dataset]
        
        fig, axes = plt.subplots(1, 2, figsize=(18, 5))

        ax1 = axes[0]
        ax1.set_yscale('log')
        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['time']
            if len(algo_data) > 0:
                sns.kdeplot(algo_data, ax=ax1, label=algo, fill=True, alpha=0.3)
        
        ax1.set_title(f'KDE of Runtime for {dataset}')
        ax1.set_xlabel('Runtime (seconds)')
        ax1.set_ylabel('Frequency')
        ax1.legend()

        ax2 = axes[1]
        plot_data = []
        labels = []

        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['time']
            if len(algo_data) > 0:
                plot_data.append(algo_data)
                labels.append(algo)
            else:
                print(f"Algorithm: {algo}, Dataset: {dataset}, Count: {len(algo_data)} - No data")

        ax2.boxplot(plot_data, tick_labels=labels, showmeans=True)
        ax2.set_title(f'Box Plot of Runtime for {dataset}')
        ax2.set_ylabel('Runtime (seconds)')
        ax2.set_yscale('log')
        plt.setp(ax2.get_xticklabels(), rotation=45, ha='right')

        plt.tight_layout()
        save_file = os.path.join(save_path, f"{dataset}_runtime_distributions.png")
        plt.savefig(save_file)
        plt.close(fig)


def plot_cost_err_distributions(datasets, results_df, algorithms, save_path):
    os.makedirs(save_path, exist_ok=True)

    for dataset in datasets:
        dataset_data = results_df[results_df['dataset'] == dataset]

        fig, axes = plt.subplots(1, 2, figsize=(18, 5))

        ax1 = axes[0]
        ax1.set_yscale('log')
        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['cost_rerr']
            if len(algo_data) > 1 and algo_data.nunique() > 1:
                sns.kdeplot(algo_data, ax=ax1, label=algo, fill=True, alpha=0.3)
            elif algo_data.nunique() == 1:
                print(f"Algorithm: {algo}, Dataset: {dataset}, Count: {len(algo_data)} - Constant value")
                const_val = algo_data.iloc[0]
                ax1.axvline(const_val, label=f"{algo} (constant)", linestyle='--')

        ax1.set_title(f'KDE of Cost Error for {dataset}')
        ax1.set_xlabel('Cost Error')
        ax1.set_ylabel('Frequency')
        ax1.legend()

        ax2 = axes[1]
        plot_data = []
        labels = []

        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['cost_rerr']
            if len(algo_data) > 0:
                plot_data.append(algo_data)
                labels.append(algo)

        ax2.boxplot(plot_data, tick_labels=labels, showmeans=True)
        ax2.set_title(f'Box Plot of Cost Error for {dataset}')
        ax2.set_ylabel('Cost Error')
        ax2.set_yscale('log')
        plt.setp(ax2.get_xticklabels(), rotation=45, ha='right')

        plt.tight_layout()
        plt.savefig(os.path.join(save_path, f"{dataset}_cost_err_distributions.png"))
        plt.close(fig)


def plot_coupling_err_distributions(datasets, results_df, algorithms, save_path):
    os.makedirs(save_path, exist_ok=True)

    for dataset in datasets:
        dataset_data = results_df[results_df['dataset'] == dataset]

        fig, axes = plt.subplots(1, 2, figsize=(18, 5))

        ax1 = axes[0]
        ax1.set_yscale('log')
        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['coupling_avg_err']
            if len(algo_data) > 0:
                sns.kdeplot(algo_data, ax=ax1, label=algo, fill=True, alpha=0.3)

        ax1.set_title(f'KDE of Coupling Avg. Error for {dataset}')
        ax1.set_xlabel('Coupling Avg. Error')
        ax1.set_ylabel('Frequency')
        ax1.legend()

        ax2 = axes[1]
        plot_data = []
        labels = []

        for algo in algorithms:
            algo_data = dataset_data[dataset_data['name'] == algo]['coupling_avg_err']
            if len(algo_data) > 0:
                plot_data.append(algo_data)
                labels.append(algo)

        ax2.boxplot(plot_data, tick_labels=labels, showmeans=True)
        ax2.set_title(f'Box Plot of Coupling Avg. Error for {dataset}')
        ax2.set_ylabel('Coupling Avg. Error')
        ax2.set_yscale('log')
        plt.setp(ax2.get_xticklabels(), rotation=45, ha='right')

        plt.tight_layout()
        plt.savefig(os.path.join(save_path, f"{dataset}_coupling_err_distributions.png"))
        plt.close(fig)


for group_name, datasets in dataset_groups.items():
    print(f"\n### {group_name} Runtime Analysis ###")
    plot_runtime_distributions(datasets, results_df, algorithms, save_path='runtime_distributions')
    print(f"\n### {group_name} Cost Error Analysis ###")
    plot_cost_err_distributions(datasets, results_df, algorithms, save_path='cost_err_distributions')
    print(f"\n### {group_name} Coupling Error Analysis ###")
    plot_coupling_err_distributions(datasets, results_df, algorithms, save_path='coupling_err_distributions')

import os
import matplotlib.pyplot as plt
from IPython.display import HTML
import dataframe_image as dfi

def save_analysis_tables_by_dimension(results_df, metrics, output_dir="analysis_tables"):
    os.makedirs(output_dir, exist_ok=True)
    
    for group_name, datasets in dataset_groups.items():
        dim_folder = os.path.join(output_dir, group_name.replace(" ", "_"))
        os.makedirs(dim_folder, exist_ok=True)
        
        dim_data = results_df[results_df['dataset'].isin(datasets)]
        
        for metric in metrics:
            mean_table = get_mean_comparison_table(dim_data, metric)
            mean_file = os.path.join(dim_folder, f"{metric}_mean.csv")
            mean_table.to_csv(mean_file, index=False)
            
            std_table = get_std_comparison_table(dim_data, metric)
            std_file = os.path.join(dim_folder, f"{metric}_std.csv")
            std_table.to_csv(std_file, index=False)
            
            def highlight_top_two(s):
                is_first = s == s.min()
                is_second = s == s[s != s.min()].min() if len(s[s != s.min()]) > 0 else pd.Series(False, index=s.index)
                return [
                    'background-color: lightgreen' if f else
                    'background-color: lightblue' if s_ else
                    ''
                    for f, s_ in zip(is_first, is_second)
                ]
            
            mean_styled = mean_table.style.apply(highlight_top_two, axis=1, subset=mean_table.columns[1:])
            mean_html_file = os.path.join(dim_folder, f"{metric}_mean.html")
            with open(mean_html_file, 'w') as f:
                f.write(mean_styled.to_html())
            
            mean_img_file = os.path.join(dim_folder, f"{metric}_mean.png")
            dfi.export(mean_styled, mean_img_file)
                
            std_styled = std_table.style.apply(highlight_top_two, axis=1, subset=std_table.columns[1:])
            std_html_file = os.path.join(dim_folder, f"{metric}_std.html")
            with open(std_html_file, 'w') as f:
                f.write(std_styled.to_html())
            
            std_img_file = os.path.join(dim_folder, f"{metric}_std.png")
            dfi.export(std_styled, std_img_file)
                
        print(f"Saved analysis tables for {group_name} to {dim_folder}")

save_analysis_tables_by_dimension(results_df, metrics_to_analyze)


import jax.numpy as jnp
import jax
from jax import jit
import ot
import ot.plot
from jax.scipy.stats import norm
import numpy as np
import matplotlib.pyplot as plt
from tqdm import trange
from typing import NamedTuple
jax.config.update("jax_enable_x64", True)

class State(NamedTuple):
    P_k: jax.Array
    lam_k: jax.Array
    eta_k: jax.Array
    computed_marginals: tuple
    k: int
    done: bool

@jax.jit
def prox_quadratic_ot(
    Pk,    # current primal iterate (n x m)
    lam,   # dual vector for row constraints (size n)
    eta,   # dual vector for col constraints (size m)
    C,     # cost matrix (n x m)
    tau,   # primal stepsize
    eps,   # regularization weight
):
    """
    Primal update for the PDHG iteration with a quadratic penalty:
      min_{P>=0} <P, C> + (eps/2)||P||^2 + (1/(2tau))||P - [Pk - tau K^T tilde{y}^k]||^2
    """

    gradient = C + lam[:, None] + eta[None, :]

    denom = 1 + tau * eps
    P_unconstrained = (Pk - tau * gradient) / denom

    return jnp.clip(P_unconstrained, 0.0, jnp.inf)


@jax.jit
def pdhg_entropic_ot_step(
    Pk, lam, eta, computed_marginals_prev,
    C, alpha, beta,
    tau, sigma, eps,
):
    """
    One iteration of PDHG for entropic OT.
      1) P^{k+1} = prox_{tau*f}( P^k - tau * K^T tilde{y}^k )
      2) y^{k+1} = y^k + sigma * K(2P^{k+1} - P^k)
      3) store old y for next iteration
    Returns: (P^{k+1}, lam^{k+1}, eta^{k+1})
    """
    #=== 1) Primal update
    P_next = prox_quadratic_ot(Pk, lam, eta, C, tau, eps)

    #=== 2) Dual update

    # Previous update was less optimized
    # P_diff = 2.0*P_next - Pk

    computed_marginals_next = (
        jnp.sum(P_next, axis=1),
        jnp.sum(P_next, axis=0)
    )

    lam_next = lam + sigma*(2*computed_marginals_next[0] - computed_marginals_prev[0]  - alpha)
    eta_next = eta + sigma*(2*computed_marginals_next[1] - computed_marginals_prev[1] - beta)

    return P_next, lam_next, eta_next, computed_marginals_next



def pdhg_quadratic_ot(
    C, alpha, beta, eps=1e-2,
    tau=0.9, sigma=0.9, tol=1e-8,
    max_outer_iter=1000, initial_point=None, save_iters=True
):
    """
    Run PDHG to solve entropic OT:
       min_{P>=0} <P,C> + eps * sum_ij P_ij log(P_ij)
       s.t. P 1 = alpha, P^T 1 = beta.
    """
    n, m = C.shape
    # Initialize primal/dual
    Pk      = jnp.ones((n,m)) * (1.0/(n*m)) if initial_point is None else initial_point[0]
    lam     = jnp.zeros(n) if initial_point is None else initial_point[1][:n]
    eta     = jnp.zeros(m) if initial_point is None else initial_point[1][n:]

    computed_marginals = (
        jnp.sum(Pk, axis=1),
        jnp.sum(Pk, axis=0)
    )

    alpha = jnp.asarray(alpha)
    beta  = jnp.asarray(beta)

    @jax.jit
    def stopping_criteria(computed_marginals):
        marginal_1, marginal_2 = computed_marginals
        marginals_error = jnp.maximum(
            jnp.linalg.norm(marginal_1 - alpha),
            jnp.linalg.norm(marginal_2 - beta)
        )
        return tol > marginals_error

    def one_step(carry):
        (P_k, lam_k, eta_k, computed_marginals, k, done) = carry
        P_next, lam_next, eta_next, computed_marginals_next = pdhg_entropic_ot_step(
            P_k, lam_k, eta_k, computed_marginals, C, alpha, beta, tau, sigma, eps,
        )
        done = stopping_criteria(computed_marginals)
        saved_iter = P_next if save_iters else None
        return State(P_next, lam_next, eta_next, computed_marginals_next, k+1, done), saved_iter


    def skip_updates(state):
        saved_iter = state.P_k if save_iters else None
        return state, saved_iter

    def outer_body_cond(state, _):
        return jax.lax.cond (
            state.done,
            skip_updates,
            one_step,
            state
        )

    # Run the iterations
    (P_final, lam_final, eta_final, _, k, done), iters = jax.lax.scan(
        outer_body_cond,
        init=State(Pk, lam, eta, computed_marginals, 0, False),
        xs=None,
        length=max_outer_iter
    )
    # if done:
    #     jax.debug.print("Algorithm converged at {} total iterations", k)
    # else:
    #     jax.debug.print("Algorithm didn't converge, so stopped after {} total iterations", k)

    return P_final, lam_final, eta_final, iters


def pdhg_algorithm(a: jnp.ndarray, b: jnp.ndarray, C: jnp.ndarray):
    P, *_ = pdhg_quadratic_ot(alpha=a, beta=b, C = C, eps=0)
    return P, jnp.sum(P * C)



import jax.numpy as jnp


def regularize_input(a: jnp.ndarray, 
                     b: jnp.ndarray, 
                     C: jnp.ndarray, 
                     max_mass_add: float = 1e-30):

    a_zero_size = jnp.sum(a == 0)
    a = a.at[a == 0].set(max_mass_add / a_zero_size)

    b_zero_size = jnp.sum(b == 0)
    b = b.at[b == 0].set(max_mass_add / b_zero_size)

    C_zero_size = jnp.sum(C == 0)
    C = C.at[C == 0].set(max_mass_add / C_zero_size)

    return a, b, C







import ot
import gc
import os
import os.path
import numpy as np
import pandas as pd
import itertools as it
import jax.numpy as jnp
import open3d as o3d
from functools import partial
from uot.core.dataset import Measure, generate_coefficients, generate_measures, get_grids, Measure
from tqdm import tqdm


def get_q_const(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    n = x.shape[0]
    m = y.shape[0]

    return ot.dist(x.reshape((n, -1)), y.reshape((m, -1)))

def get_exact_solution(a: np.ndarray, b: np.ndarray, C: np.ndarray) -> tuple[np.ndarray, float]:
    T = ot.emd(a, b, C, numItermax=10000000, numThreads=os.cpu_count())
    return T, np.sum(T * C)

def generate_two_fold_problems(grid, measures: list[Measure], name: str, one_cost=False):
    ot_problems = []
    if one_cost:
        source_points, _ = measures[0].to_histogram()
        target_points, _ = measures[1].to_histogram()
        C = get_q_const(source_points, target_points)
    else:
        C = get_q_const
    for source_measure, target_measure in it.combinations(measures, 2):
        ot_problem = OTProblem(name="Simple transport", 
                                     source_measure=source_measure,
                                     target_measure=target_measure,
                                     C=C)
        ot_problem.kwargs.update({'dataset': name})
        ot_problems.append(ot_problem)
    return ot_problems

def generate_data_problems(data_type: str, num_points: int, num_samples: int = 10):
    """
    Generates OT problems from data files in the Data folder.
    
    Args:
        data_type (str): The type of data to use (e.g., 'CauchyDensity', 'ClassicImages')
        num_points (int): The number of points/resolution (e.g., 32, 64, 128, 256, 512)
        num_samples (int): Maximum number of samples to use (default: 10)
        create_grids (bool): Whether to create grid points for the measures (default: True)
    
    Returns:
        list[OTProblem]: A list of OTProblem objects created from the data

    Full list of data types:
        - WhiteNoise
        - Cauchy density
        - GRFmoderate
        - GRFrough
        - GRFsmooth
        - LogGRF
        - LogitGRF
        - MicroscopyImages
        - Shapes
        - ClassicImages
    """
    data_folder = os.path.join("datasets", "DOTmark_1.0", "Data", data_type)
    
    if not os.path.exists(data_folder):
        raise ValueError(f"Data folder '{data_folder}' does not exist")
    
    file_pattern = f"data{num_points}_"
    data_files = [f for f in os.listdir(data_folder) if file_pattern in f]
    
    if len(data_files) > num_samples:
        data_files = data_files[:num_samples]
    
    measures = []
    for file_name in data_files:
        file_path = os.path.join(data_folder, file_name)
        data = pd.read_csv(file_path, header=None).values
        
        data = data / data.sum()
        
        x = np.linspace(0, 1, data.shape[1])
        y = np.linspace(0, 1, data.shape[0])
        grid = np.meshgrid(x, y)
        
        measure = Measure(
            name=f"{data_type}_{file_name.replace('.csv', '')}",
            support=grid,
            distribution=data,
            kwargs={"data_type": data_type, "file_name": file_name}
        )
        
        measures.append(measure)
    
    problems = []
    for source_measure, target_measure in it.combinations(measures, 2):
        ot_problem = OTProblem(
            name=f"{data_type} {num_points}x{num_points}",
            source_measure=source_measure,
            target_measure=target_measure,
            C=get_q_const
        )
        ot_problem.kwargs.update({'dataset': f"{data_type} {num_points}x{num_points}"})
        ot_problem.kwargs.update({"data_type": data_type, "num_points": num_points})
        problems.append(ot_problem)
    
    return problems

def generate_3d_mesh_problems(num_points: int = None, num_meshes: int = 10, color_mode: str = "r"):
    """
    Generates OT problems from 3D colored mesh files in the Reference_3D_colored_meshes folder.
    
    Args:
        num_points (int, optional): The number of points to sample from each 3D mesh. 
                                   If None, uses the actual number of vertices in each mesh.
        num_meshes (int): Maximum number of mesh files to use (default: 10)
        color_mode (str): How to handle color channels:
                          - "r", "g", "b": Use a single color channel
                          - "separate": Create separate problems for each color channel
    
    Returns:
        list[OTProblem]: A list of OTProblem objects created from the 3D meshes
    """
    mesh_folder = os.path.join("datasets", "color_meshes")
    
    if not os.path.exists(mesh_folder):
        raise ValueError(f"Mesh folder '{mesh_folder}' does not exist")
    
    mesh_files = [f for f in os.listdir(mesh_folder) if f.endswith('.ply')]
    
    if len(mesh_files) > num_meshes:
        mesh_files = mesh_files[:num_meshes]
    
    if color_mode.lower() == "r":
        channels = [0]
        channel_names = ["red"]
    elif color_mode.lower() == "g":
        channels = [1]
        channel_names = ["green"]
    elif color_mode.lower() == "b":
        channels = [2]
        channel_names = ["blue"]
    elif color_mode.lower() == "separate":
        channels = [0, 1, 2]
        channel_names = ["red", "green", "blue"]
    else:
        raise ValueError(f"Invalid color_mode: {color_mode}. Must be 'r', 'g', 'b', or 'separate'")
    
    measures_by_channel = {channel_name: [] for channel_name in channel_names}
    
    for file_name in mesh_files:
        file_path = os.path.join(mesh_folder, file_name)
        
        try:
            mesh = o3d.io.read_triangle_mesh(file_path)
            
            mesh_num_points = len(np.asarray(mesh.vertices))
            sampling_points = num_points if num_points is not None else mesh_num_points
            
            if sampling_points > mesh_num_points:
                print(f"Warning: Requested {sampling_points} points but {file_name} only has {mesh_num_points} vertices. Using all vertices.")
                sampled_points = mesh
            else:
                sampled_points = mesh.sample_points_uniformly(sampling_points)
            
            points = np.asarray(sampled_points.points)
            colors = np.asarray(sampled_points.colors)
            
            for channel, channel_name in zip(channels, channel_names):
                distribution = colors[:, channel]
                distribution = distribution / distribution.sum()
            
                measure = Measure(
                    name=f"3DMesh_{file_name.replace('.ply', '')}_{channel_name}",
                    support=[points[:, 0], points[:, 1], points[:, 2]],
                    distribution=distribution,
                    kwargs={
                        "mesh_name": file_name, 
                        "color_channel": channel,
                        "color_name": channel_name,
                        "num_points": len(points)
                    }
                )
                measures_by_channel[channel_name].append(measure)
            
        except Exception as e:
            print(f"Error loading {file_name}: {e}")
    

    all_problems = []
    
    for channel_name, measures in measures_by_channel.items():
        for source_measure, target_measure in it.combinations(measures, 2):
            source_points = source_measure.kwargs.get("num_points")
            target_points = target_measure.kwargs.get("num_points")
            
            ot_problem = OTProblem(
                name=f"3D_Colored_Mesh_{channel_name}_{source_points}x{target_points}pts",
                source_measure=source_measure,
                target_measure=target_measure,
                C=get_q_const
            )
            ot_problem.kwargs.update({
                'dataset': f"3D_Colored_Mesh_{channel_name}_{source_points}x{target_points}pts",
                "data_type": "3D_Colored_Mesh", 
                "source_points": source_points,
                "target_points": target_points,
                "color_mode": channel_name
            })
            all_problems.append(ot_problem)
    
    return all_problems

def get_distribution_problemset(name: str, coeffs=None):
    size, dim_str, distributions_str = name.split(' ')
    distributions = sorted(distributions_str.split('|'))
    dim = int(dim_str[0])

    if dim not in [1, 2, 3]:
        raise ValueError(f"Invalid dimension: {dim}. Expected 1, 2, or 3.")

    size_parts = size.split('x')
    if len(size_parts) != dim:
        raise ValueError(f"Invalid size format: {size}. Expected format: {'x'.join(['<size>'] * dim)}.")
    
    grid_size = int(size_parts[0])
    
    distribution_counts = {distribution: 10 // len(distributions) for distribution in distributions}
    
    if coeffs is None:
        coeffs = generate_coefficients(dim, distribution_counts)
    
    grids = get_grids(dim, [grid_size], start=-10, end=10)
    measures = generate_measures(dim, coeffs, grids)
    
    measure_key = f"{'x'.join([str(grid_size)] * dim)} {dim}D {'|'.join(distributions)}"
    
    try:
        problems = generate_two_fold_problems(None, measures[measure_key], name=measure_key)
        return problems
    except KeyError as e:
        available_keys = list(measures.keys())
        raise KeyError(f"Key '{measure_key}' not found in measures. Available keys: {available_keys}")

def get_problemset(problem_spec, coeffs=None, **kwargs):
    if isinstance(problem_spec, str):
        return get_distribution_problemset(problem_spec, coeffs)
    
    if not isinstance(problem_spec, tuple) or len(problem_spec) < 3:
        raise ValueError("Problem spec must be either a string or a tuple (type, name, num_points[, dims])")
    
    dimensionality, name, num_points = problem_spec[:3]
    dims = problem_spec[3] if len(problem_spec) > 3 else 1
    
    if dimensionality == 1:
        if dims == 1:
            size_str = f"{num_points}"
        elif dims == 2:
            size_str = f"{num_points}x{num_points}"
        elif dims == 3:
            size_str = f"{num_points}x{num_points}x{num_points}"
        else:
            raise ValueError(f"Invalid dimension: {dims}. Expected 1, 2, or 3.")
        
        problem_str = f"{size_str} {dims}D {name}"
        return get_distribution_problemset(problem_str, coeffs)

    elif dimensionality == 2:
        data_type = name
        num_samples = kwargs.get("num_samples", 10)
        return generate_data_problems(data_type=data_type, num_points=num_points, num_samples=num_samples)
    
    elif dimensionality == 3:
        color_mode = name
        num_meshes = kwargs.get("num_meshes", 10)
        return generate_3d_mesh_problems(num_points=num_points, color_mode='r', num_meshes=num_meshes)
   
    else:
        raise ValueError(f"Unknown problem type: {dimensionality}. Expected 'distribution', '3d_mesh', or 'data'")

def run_experiment(experiment: 'Experiment',
                   jit_algorithms=None,
                   solvers: dict[str, callable] = None,
                   problemsets_names: list[tuple] = None,
                   folds: int = 1) -> pd.DataFrame:
    """
    Executes a series of experiments using specified solvers on a set of problems.
    Args:
        experiment (Experiment): The experiment object that defines how to run the experiments.
        solvers (dict[str, callable]): A dictionary where keys are solver names and values are tuples 
            containing the solver function and optional sets of keyword arguments for the solver.
        problemsets_names (list[tuple]): A list of problem set names to retrieve and use in the experiments.
        jit_algorithms (list[str], optional): A list of algorithm names to exclude from the results. Defaults to None.
        folds (int, optional): The number of folds for cross-validation. Defaults to 1.
    Returns:
        pd.DataFrame: A DataFrame containing the results of the experiments, including solver names 
        and any additional parameters used.
    """

    problem_sets = [get_problemset(name) for name in problemsets_names]
    problems = [problem for problemset in problem_sets for problem in problemset]

    problems *= folds

    for problem in problems:
        source_distribution = problem.source_measure.distribution
        target_distribution = problem.target_measure.distribution
        
        if np.any(np.logical_or(np.isnan(source_distribution), np.isinf(source_distribution))):
            print(f"Detected problem in: {problem}")
            print("Source distribution contains NaN or Inf")
            print(source_distribution)
            raise ValueError("Nan or Inf in source distributions")
        
        if np.any(np.logical_or(np.isnan(target_distribution), np.isinf(target_distribution))):
            print(f"Detected problem in: {problem}")
            print("Source distribution contains NaN or Inf")
            print(target_distribution)
            raise ValueError("Nan or Inf in target distributions")
    
    dfs = []

    solvers_number = sum(len(kwargs) if kwargs else 1 for _, kwargs in solvers.values())

    with tqdm(total=solvers_number * len(problems), desc="Running experiments") as pbar:
        progress_callback = lambda: pbar.update(1)

        for solver_name, solver in solvers.items():
            
            solver_function, kwargs_sets = solver
            kwargs_sets = kwargs_sets if kwargs_sets else [{}]
            solvers = [(partial(solver_function, **kwargs), kwargs) for kwargs in kwargs_sets]

            for solver, kwargs in solvers:
                pbar.set_description(f"Solver: {solver_name}({kwargs})")
                solver_result = experiment.run_experiment(ot_problems=problems, progress_callback=progress_callback, solver=solver)

                solver_result['name'] = solver_name
                
                for kwarg_name, value in kwargs.items():
                    solver_result[kwarg_name] = value
            
            dfs.append(solver_result)

    df = pd.concat(dfs)

    for dataset in df.dataset.unique():
        for algorithm_name in jit_algorithms:
            algorithm_results = df[(df.dataset == dataset) & (df.name == algorithm_name)]
            if len(algorithm_results):
                df.drop(algorithm_results.index[0], inplace=True)

    return df


class OTProblem:

    def __init__(self, name: str, source_measure: Measure, target_measure: Measure, C: np.ndarray, kwargs: dict = None):
        self.name = name
        self.source_measure = source_measure
        self.target_measure = target_measure

        self._C_cache = None
        self._C = C

        self._exact_cost = None
        self._exact_map = None 

        self.kwargs = kwargs if kwargs is not None else {}
    
    @property
    def C(self):
        if callable(self._C) and self._C_cache is None:
            self._C_cache = self._C(self.source_measure.get_flat_support(),
                               self.target_measure.get_flat_support())
            self._C_cache /= self._C_cache.max()
            return self._C_cache
        return self._C_cache
    
    @property
    def a(self):
        return self.source_measure.to_histogram()[1]
    
    @property
    def b(self):
        return self.target_measure.to_histogram()[1]

    def free_memory(self):
        del self._C_cache
        self._C_cache = None
        self._exact_cost = None
        self._exact_map = None

    def to_jax_arrays(self):
        a = jnp.array(self.a)
        b = jnp.array(self.b)
        C = jnp.array(self.C)

        return a, b, C

    @property
    def exact_cost(self):
        if self._exact_cost is None:
            self._exact_map, self._exact_cost = get_exact_solution(self.a, self.b, self.C)
        return self._exact_cost

    @property
    def exact_map(self):
        if self._exact_map is None:
            self._exact_map, self._exact_cost = get_exact_solution(self.a, self.b, self.C)
        return self._exact_map

    def __hash__(self):
        return hash(self.name) + hash(str(self.source_measure.kwargs)) + hash(str(self.target_measure.kwargs))
        
    def __eq__(self, other):
        if not isinstance(other, OTProblem):
            return False
        
        return self.name == other.name and \
               self.source_measure == other.source_measure and \
               self.target_measure == other.target_measure and \
               np.array_equal(self.C, other.C) and \
               self.kwargs == other.kwargs

    def __str__(self):
        return f"<OTProblem: {self.name} source={self.source_measure}, target={self.target_measure}>"

    def to_dict(self) -> dict:
        problem_dict = {'source_measure_name': self.source_measure.name, 'target_measure_name': self.target_measure.name}
        source_kwargs = { f"source_{key}": value for key, value in self.source_measure.kwargs.items() }
        target_kwargs = { f"target_{key}": value for key, value in self.target_measure.kwargs.items() }
        problem_dict.update(source_kwargs)
        problem_dict.update(target_kwargs)
        problem_dict.update(self.kwargs)
        return problem_dict


class Experiment:

    def __init__(self, name: str, run_function: callable):
        self.name = name
        self.run_function = run_function

    def run_experiment(self, ot_problems: list[OTProblem], progress_callback: callable = None, solver = None) -> dict:
        results = {}
        for i, ot_problem in enumerate(ot_problems):

            run_function = partial(self.run_function, solver=solver)

            results[ot_problem] = run_function(ot_problem)
            if progress_callback is not None:
                progress_callback()
            ot_problem.free_memory()
            if i % 100 == 0:
                gc.collect()

        df_rows = []
        for ot_problem in ot_problems:
            row_dict = ot_problem.to_dict() 
            row_dict.update(results[ot_problem])
            df_rows.append(row_dict)

        return pd.DataFrame(df_rows)
    
    def run_single(self, ot_problem: OTProblem) -> dict:
        return self.run_function(ot_problem)



import jax
import numpy as np
import jax.numpy as jnp
from jaxopt import LBFGS

    
@jax.jit
def lbfgs_multimarginal(marginals: jnp.ndarray,
             C: jnp.ndarray,
             epsilon: float = 1,
             tolerance: float = 1e-4):

    N = marginals.shape[0]
    n = marginals.shape[1]

    shapes = [tuple(n if j == i else 1 for j in range(N)) for i in range(N)]
    potentials = jnp.zeros_like(marginals)

    @jax.jit
    def objective(potentials: jax.Array):
        """Computes the dual objective with logsumexp stabilization."""
        potentials_reshaped = [potentials[i].reshape(shapes[i]) for i in range(N)]
        potentials_sum = sum(potentials_reshaped)
        log_sub_entropy = (potentials_sum - C) / epsilon
        max_log_sub_entropy = jnp.max(log_sub_entropy, axis=0, keepdims=True)
        stable_sum = jnp.exp(max_log_sub_entropy) * jnp.sum(
            jnp.exp(log_sub_entropy - max_log_sub_entropy), axis=0
        )
        dual = potentials * marginals
        return -jnp.sum(dual - epsilon * stable_sum)

    solver = LBFGS(fun=objective, tol=tolerance)
    result = solver.run(init_params=potentials)

    return result.params, result.state.error < tolerance

def lbfgs_ot(a: jnp.ndarray, 
             b: jnp.ndarray,
             C: jnp.ndarray,
             tolerance: float = 1e-4, 
             epsilon: float = 1e-3):
    marginals = jnp.array([a, b])
    potentials, converged = lbfgs_multimarginal(marginals=marginals,
                                                C = C,
                                                tolerance=tolerance,
                                                epsilon=epsilon)

    P = jnp.exp(
        (potentials[0][None, :] + potentials[1][:, None] - C) / epsilon
    )

    return P, jnp.sum(P * C), converged